# core/dispatch.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2014 B.H.J Thate"

## IMPORTS

from core.utils import get_plugname, get_name, get_strace
from core.object import Object

import collections
import threading
import logging
import queue
import time

## RUNNER

class Runner(threading.Thread):

    """ the working unit of CORED.  """

    _nr = 0

    def __init__(zelf, *args, **kwargs):
        threading.Thread.__init__(zelf, None, zelf._loop, "Runner-%s" % str(zelf._nr), args, kwargs)
        zelf.setDaemon(True)
        zelf._queue = queue.Queue()
        zelf._ready = threading.Event() 
        zelf._status = "starting"
        zelf._nr += 1

    def _loop(zelf):
        """ loop that executes the (function, object) pairs. """
        zelf._status = "waiting"
        while zelf._status:
            args, kwargs = zelf._queue.get()
            if not zelf._status: break
            func = args[0]
            zelf.setName(get_name(func))
            try: obj = args[1]
            except: obj = Object()
            if not zelf._status: obj.ready() ; break
            zelf._status = str(func)
            logging.warn("func %s" % get_name(func))
            func(*args[1:], **kwargs)
            zelf._status = "waiting"
            obj.ready()
            time.sleep(0.01)

    def exit(zelf, *args, **kwargs):
        """ stop the Runner. """
        zelf._status = ""
        zelf.put()

    def put(zelf, *args, **kwargs):
        """ put arguments/kwargs to the Runner. """
        zelf._queue.put_nowait((args, kwargs))

    def show(zelf):
        """ list of key,value pairs. """
        return ["%s=%s" % (a, zelf._state[a]) for a in zelf._state.keys()]

## DISPATCH

class Dispatcher(Object):

    """ the dispatcher delegates the workload to the Runners. Runner gets instantiated when needed. """

    def __init__(zelf, *args, **kwargs):
        Object.__init__(zelf, *args, **kwargs)
        zelf._runners = collections.deque()
        zelf._status = "starting"
        zelf._max = 50

    def is_alive(zelf, *args, **kwargs):
        """ whether any Runners running. """
        for runner in zelf._runners:
           if runner.is_alive(): return True
        return False

    def register(zelf, *args, **kwargs):
        """ callback type with corresponding callback function. """
        o = Object()
        o.cmnd = args[0]
        o.func = args[1]
        o.plugname = get_plugname()
        if o.cmnd not in zelf: zelf[o.cmnd] = []
        zelf[o.cmnd].append(o)

    def denied(zelf, *args, **kwargs):
        for obj in zelf.objects("denied"):
            if args[0] == obj.get("denied", None): return True
        return False

    def allowed(zelf, *args, **kwargs):
        for obj in zelf.objects("allowed"):
            if args[0] == obj.get("allowed", None): return True
        return False
 
    def dispatch(zelf, *args, **kwargs):
        """ dispatch an event onto a Runner. """
        event = args[0]
        zelf.execute(event.get_cmnd(), event)
        return event

    def execute(zelf, *args, **kwargs):
        """ execute a command, event pair. """
        cmnd = args[0]
        event = args[1]
        if "origin" in event:
            if zelf.allowed(event.origin): 
                logging.warn("! allow %s" % event.origin)
            elif zelf.denied(event.origin):
                logging.warn("! deny %s" % event.origin)
                return
        for functor in zelf.get(cmnd, []):
            logging.info("> %s" % get_name(functor.func))
            functor.func(event)
        return event

    def run_func(zelf, *args, **kwargs):
        """ execute command/event pair. """
        cmnd = args[0]
        event = args[1]
        if cmnd not in zelf: return
        for functor in zelf[cmnd]: functor.func(event)
        event.ready()
        return event

    def exit(zelf, name=None):
        """ stop the Runners. """
        for runner in zelf._runners:
            logging.warn("join %s" % str(runner))
            runner.join(0.2)

    def put(zelf, *args, **kwargs):
        """ put load to the Runner. """
        logging.info("> put %s" % get_name(*args))
        zelf.cleanup()
        target = zelf.make_new()
        target.put(*args, **kwargs)
        return target

    def make_new(zelf, *args, **kwargs):
        """ create a Runner. """
        for runner in zelf._runners:
            if runner._status == "waiting": return runner
        if len(zelf._runners) < zelf._max:
            runner = Runner(*args, **kwargs)
            zelf._runners.append(runner)
            runner.start()
        else: runner = random.choice(zelf._runners)
        return runner

    def cleanup(zelf, dojoin=False):
        """ remove idle Runners. """
        todo = []
        for runner in zelf._runners:
            if runner._status == "waiting": todo.append(runner)
        for runner in todo:
            runner.exit()
            runner.join(0.1)
            zelf._runners.remove(runner)
